#!/usr/bin/env python3
# ==============================================================================
#
#          FILE:  geojsonconcat.py
#
#         USAGE:  geojsonconcat --help
#                 ./src/gisconflation/geojsonconcat.py --help
#
#   DESCRIPTION:  ---
#
#       OPTIONS:  ---
#
#  REQUIREMENTS:  - python3
#          BUGS:  ---
#         NOTES:  ---
#       AUTHORS:  Emerson Rocha <rocha[at]ieee.org>
# COLLABORATORS:  ---
#
#       COMPANY:  EticaAI
#       LICENSE:  GNU Affero General Public License v3.0 or later
#                 SPDX-License-Identifier: AGPL-3.0-or-later
#       VERSION:  v1.0.0
#       CREATED:  2023-05-31 11:23 BRT
# ==============================================================================

import argparse
import csv

# import csv
# import dataclasses
import json

# import re
import sys

__VERSION__ = "1.0.0"

PROGRAM = "geojsonconcat"

# https://www.rfc-editor.org/rfc/rfc7946
# https://www.rfc-editor.org/rfc/rfc8142.html
DESCRIPTION = """
------------------------------------------------------------------------------
Concat 2 or more GeoJSON files into GeoJSON (RFC 7946) or
GeoJSON Text Sequences (RFC 8142)

@TODO for input files know to be GeoJSONSeq, read line by line

------------------------------------------------------------------------------
""".format(
    __file__
)

# https://www.rfc-editor.org/rfc/rfc7946
# The GeoJSON Format
# https://www.rfc-editor.org/rfc/rfc8142
# GeoJSON Text Sequences

# __EPILOGUM__ = ""
__EPILOGUM__ = """
------------------------------------------------------------------------------
                            EXEMPLŌRUM GRATIĀ
------------------------------------------------------------------------------
    {0} --help

    {0} input-a.geojson input-b.geojson > merged-a+b.geojson

    {0} --format-output=RFC8142 input-a.geojson input-b.geojson \
> merged-a+b.geojsonseq

------------------------------------------------------------------------------
                            EXEMPLŌRUM GRATIĀ
------------------------------------------------------------------------------
""".format(
    PROGRAM
)

STDIN = sys.stdin.buffer

# MATCH_EXACT = 1
# MATCH_NEAR = 3


class Cli:
    """Main CLI parser"""

    def __init__(self):
        """
        Constructs all the necessary attributes for the Cli object.
        """
        self.pyargs = None
        self.EXIT_OK = 0
        self.EXIT_ERROR = 1
        self.EXIT_SYNTAX = 2

    def make_args(self):
        """make_args

        Args:
            hxl_output (bool, optional): _description_. Defaults to True.
        """
        parser = argparse.ArgumentParser(
            prog=PROGRAM,
            description=DESCRIPTION,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            epilog=__EPILOGUM__,
        )

        parser.add_argument(
            "inputs", help="One or more input GeoJSON/GeoJSONSeq", nargs="+"
        )
        # parser.add_argument("geodataset_b", help="GeoJSON dataset 'B'")

        # # @TODO maybe implement GeoJSONSeq
        # parser.add_argument("input", help="Input GeoJSON / GeoJSONSeq. Use - for stdin")
        # # parser.add_argument("geodataset_b", help="GeoJSON dataset 'B'")

        # @see https://stevage.github.io/ndgeojson/
        # @see https://datatracker.ietf.org/doc/html/rfc8142
        parser.add_argument(
            "--format-output",
            help="Output format GeoJSON (RFC7946) or GeoJSON Text Sequences (RFC 8142)",
            dest="format_output",
            default="RFC7946",
            required=False,
            choices=["RFC7946", "RFC8142"],
            # choices=["auto", "geojson", "geojsonseq"],
            nargs="?",
        )

        return parser.parse_args()

    def execute_cli(self, pyargs, stdin=STDIN, stdout=sys.stdout, stderr=sys.stderr):
        gconcat = GeoJSOnConcater(
            inputs=pyargs.inputs,
            # geodataset_b=pyargs.geodataset_b,
            # skip_invalid_geometry=True,
        )
        # print("TODO")

        gconcat.output()
        # gmerger.debug()
        return self.EXIT_OK


# class GeoJSONMerger:
class GeoJSOnConcater:
    def __init__(self, inputs: list, output_format: str = None) -> None:
        self.inputs = inputs
        self.output_format = output_format

    def _loader_temp(self, input_ptr: str):
        # @TODO make this efficient for very large files
        data = None
        if input_ptr == "-":
            temp = []
            for line in sys.stdin:
                # temp.append(line.strip())
                temp.append(line.rstrip())
                # temp.append(line)

            data = "\n".join(temp)
        else:
            with open(input_ptr, "r") as _file:
                data = _file.read()
            # pass

        return data

    def debug(self):
        # print("todo")
        print(self.statistics)

    def output(self):
        # @TODO keep other metadata at top level, if exist

        if self.output_format != "RFC8142":
            print('{"type": "FeatureCollection", "features": [')

        count_files = len(self.inputs)

        for infile in self.inputs:
            # print("TODO", infile)

            # @TODO optimize if know upfront is GeoJSONseq (maybe file extension)
            infile_str = self._loader_temp(infile)
            infile_obj = json.loads(infile_str)

            if "features" not in infile_obj:
                raise SyntaxError(f"File {infile} not GeoJSON. Must have features")

            count_files -= 1
            count_items_now = len(infile_obj["features"])

            for item in infile_obj["features"]:
                count_items_now -= 1
                line_str = json.dumps(item, ensure_ascii=False)
                if self.output_format != "RFC8142" and (
                    count_items_now > 0 or count_files > 0
                ):
                    line_str += ","

                # @TODO deal with geojsonseq RS on this part

                # https://www.rfc-editor.org/rfc/rfc8142
                if self.output_format == "RFC8142":
                    print(f"\x1e{line_str}\n", sep="", end="")
                else:
                    print(line_str)

        # count = len(self.out_dict.keys())
        # for _key, item in self.out_dict.items():
        #     count -= 1
        #     line_str = json.dumps(item, ensure_ascii=False)
        #     if count > 0:
        #         line_str += ","
        #     print(line_str)
        if self.output_format != "RFC8142":
            print("]}")

        # print("")
        # self.debug()


# https://stackoverflow.com/questions/3952132/how-do-you-dynamically-identify-unknown-delimiters-in-a-data-file
def get_delimiter(file_path: str, encoding="utf-8") -> str:
    with open(file_path, "r", encoding=encoding) as csvfile:
        delimiter = str(csv.Sniffer().sniff(csvfile.read()).delimiter)
        return delimiter


def exec_from_console_scripts():
    main = Cli()
    args = main.make_args()
    main.execute_cli(args)


if __name__ == "__main__":
    main = Cli()
    args = main.make_args()
    main.execute_cli(args)



"""
# @DEPRECATED just use osmium export (pip install osmium)


@see https://github.com/osmcode/pyosmium/blob/master/examples/convert_to_geojson.py
# python scripts/etc/osm-convert_to_geojson.py
# python scripts/etc/convert_to_geojson.py data/tmp/brasil-uf.osm.pbf > data/tmp/brasil-uf.osm.geojson
# Requires:
#     pip install osmium


Output an OSM file as geojson.

This demonstrates how to use the GeoJSON factory.
"""
import sys
import json

import osmium as o

geojsonfab = o.geom.GeoJSONFactory()


class GeoJsonWriter(o.SimpleHandler):
    def __init__(self):
        super().__init__()
        # write the Geojson header
        print('{"type": "FeatureCollection", "features": [')
        self.first = True

    def finish(self):
        print("]}")

    def node(self, o):
        if o.tags:
            self.print_object(geojsonfab.create_point(o), o.tags)

    def way(self, o):
        if o.tags and not o.is_closed():
            self.print_object(geojsonfab.create_linestring(o), o.tags)

    def area(self, o):
        if o.tags:
            self.print_object(geojsonfab.create_multipolygon(o), o.tags)

    def print_object(self, geojson, tags):
        geom = json.loads(geojson)
        if geom:
            feature = {"type": "Feature", "geometry": geom, "properties": dict(tags)}
            if self.first:
                self.first = False
            else:
                print(",")

            print(json.dumps(feature))


def main(osmfile = None, is_stdin = False):
    handler = GeoJsonWriter()

    # @TODO implement from stdin

    if is_stdin:
        buff = sys.stdin.read().encode()
        handler.apply_buffer(buffer=buff)
    else:
        handler.apply_file(osmfile)
    handler.finish()

    return 0


def exec_from_console_scripts():

    print(sys.stdin.isatty())

    if not sys.stdin.isatty() and len(sys.argv) != 2:
        # print ("not sys.stdin.isatty")
        sys.exit(main(None, True))
    else:
        sys.exit(main(sys.argv[1], None))

    if len(sys.argv) != 2:
        print("Usage: python %s <osmfile>" % sys.argv[0])
        sys.exit(-1)
    sys.exit(main(sys.argv[1]))


if __name__ == "__main__":

    print(sys.stdin.isatty())
    if not sys.stdin.isatty() and len(sys.argv) != 2:
        # print ("not sys.stdin.isatty")
        sys.exit(main(None, True))
    else:
        sys.exit(main(sys.argv[1], None))


    if len(sys.argv) != 2:
        print("Usage: python %s <osmfile>" % sys.argv[0])
        sys.exit(-1)
    sys.exit(main(sys.argv[1]))



#!/usr/bin/env python3
# ==============================================================================
#
#          FILE:  csv2excel.py
#
#         USAGE:  ./scripts/csv2excel.py
#                 ./scripts/csv2excel.py --help
#
#   DESCRIPTION:  ---
#
#       OPTIONS:  ---
#
#  REQUIREMENTS:  - python3
#                   - pip install pandas (https://pandas.pydata.org/)
#          BUGS:  ---
#         NOTES:  ---
#       AUTHORS:  Emerson Rocha <rocha[at]ieee.org>
# COLLABORATORS:  ---
#
#       COMPANY:  EticaAI
#       LICENSE:  Public Domain dedication or Zero-Clause BSD
#                 SPDX-License-Identifier: Unlicense OR 0BSD
#       VERSION:  v0.2.0
#       CREATED:  2023-04-23 01:04 BRT
#      REVISION:  ---
# ==============================================================================

import sys
import argparse
import pandas as pd

__VERSION__ = "0.2.0"
PROGRAM = "csv2excel"
DESCRIPTION = """
------------------------------------------------------------------------------
{0} v{1} Convert CSV to Excel file with autofilters

------------------------------------------------------------------------------
""".format(
    PROGRAM, __VERSION__
)

__EPILOGUM__ = """
------------------------------------------------------------------------------
                            EXEMPLŌRUM GRATIĀ
------------------------------------------------------------------------------
    {0} input.csv output.xlsx
------------------------------------------------------------------------------
                            EXEMPLŌRUM GRATIĀ
------------------------------------------------------------------------------
""".format(
    PROGRAM
)

STDIN = sys.stdin.buffer


class Cli:
    def __init__(self):
        """
        Constructs all the necessary attributes for the Cli object.
        """
        self.pyargs = None
        self.EXIT_OK = 0
        self.EXIT_ERROR = 1
        self.EXIT_SYNTAX = 2

    def make_args(self):
        """make_args"""
        parser = argparse.ArgumentParser(
            prog=PROGRAM,
            description=DESCRIPTION,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            epilog=__EPILOGUM__,
        )

        parser.add_argument("input_csv", help="Input CSV file")
        parser.add_argument("output_xlsx", help="Output XLSX")

        # # @see https://stackoverflow.com/questions/41669690/how-to-overcome-the-limit-of-hyperlinks-in-excel
        # parser.add_argument(
        #     "--hiperlink",
        #     help="Add hiperlink to column names if they have non-empty value. "
        #     "alredy are mached with each other. "
        #     "Use '||' to deparate colum name from the URL. "
        #     "Add {value} as part of the URL for placeholder of external link. "
        #     "Accept multiple values. "
        #     "Example: "
        #     "--pivot-key-main='colum_name||http://example.com/page/{value}'",
        #     dest="hiperlink",
        #     nargs="?",
        #     action="append",
        # )

        return parser.parse_args()

    def execute_cli(self, pyargs, stdin=STDIN, stdout=sys.stdout, stderr=sys.stderr):
        csv2excel(pyargs.input_csv, pyargs.output_xlsx)
        return self.EXIT_OK


def csv2excel(input_csv: str, output_xlsx: str, delimiter: str = ","):
    """csv2excel
    @see https://xlsxwriter.readthedocs.io/example_pandas_autofilter.html

    Args:
        input_csv (str): Input CSV file
        output_xlsx (str): Output XLSX
    """
    # print("TODO")
    df = pd.read_csv(input_csv, sep=delimiter)

    # Create a Pandas Excel writer using XlsxWriter as the engine.
    writer = pd.ExcelWriter(output_xlsx, engine="xlsxwriter")

    # Convert the dataframe to an XlsxWriter Excel object. We also turn off the
    # index column at the left of the output dataframe.
    df.to_excel(writer, sheet_name="Sheet1", index=False)

    # Get the xlsxwriter workbook and worksheet objects.
    workbook = writer.book
    worksheet = writer.sheets["Sheet1"]

    # worksheet.write_url("A2", "ftp://www.python.org/")

    # Get the dimensions of the dataframe.
    (max_row, max_col) = df.shape

    # Make the columns wider for clarity.
    worksheet.set_column(0, max_col - 1, 12)

    # Set the autofilter.
    worksheet.autofilter(0, 0, max_row, max_col - 1)

    # Add an optional filter criteria. The placeholder "Region" in the filter
    # is ignored and can be any string that adds clarity to the expression.
    # worksheet.filter_column(0, "Region == East")

    # # It isn't enough to just apply the criteria. The rows that don't match
    # # must also be hidden. We use Pandas to figure our which rows to hide.
    # for row_num in df.index[(df["Region"] != "East")].tolist():
    #     worksheet.set_row(row_num + 1, options={"hidden": True})

    # Close the Pandas Excel writer and output the Excel file.
    writer.close()


def parse_argument_values(arguments: list, delimiter: str = "||") -> dict:
    if not arguments or len(arguments) == 0 or not arguments[0]:
        return None

    result = {}
    for item in arguments:
        if item.find(delimiter) > -1:
            _key, _val = item.split(delimiter)
            result[_key] = _val
        else:
            result[_key] = True
    return result


def exec_from_console_scripts():
    main = Cli()
    args = main.make_args()
    main.execute_cli(args)


if __name__ == "__main__":
    cli_main = Cli()
    args = cli_main.make_args()
    cli_main.execute_cli(args)



#!/usr/bin/env python3
# ==============================================================================
#
#          FILE:  dictionarybuilder.py
#
#         USAGE:  ./scripts/dictionarybuilder.py
#                 ./scripts/dictionarybuilder.py --help
#
#   DESCRIPTION:  ---
#
#       OPTIONS:  ---
#
#  REQUIREMENTS:  - python3
#          BUGS:  ---
#         NOTES:  ---
#       AUTHORS:  Emerson Rocha <rocha[at]ieee.org>
# COLLABORATORS:  ---
#
#       COMPANY:  EticaAI
#       LICENSE:  GNU Affero General Public License v3.0 or later
#                 SPDX-License-Identifier: AGPL-3.0-or-later
#       VERSION:  v1.0.0
#       CREATED:  2023-05-06 19:32 BRT
#      REVISION:  --
# ==============================================================================

import argparse
import csv
import re
import sys


__VERSION__ = "1.0.0"
PROGRAM = "dictionarybuilder"
DESCRIPTION = """
------------------------------------------------------------------------------
Convert 2 or more columns of one CSV input file into a dictionary (a sorted
2 column RS delimited file) optimized to be used by other tools as simple
pivot file to replace or append new values to existing datasets

------------------------------------------------------------------------------
""".format(
    PROGRAM, __VERSION__
)

__EPILOGUM__ = """
------------------------------------------------------------------------------
                            EXEMPLŌRUM GRATIĀ
------------------------------------------------------------------------------

    {0} --help

------------------------------------------------------------------------------
                            EXEMPLŌRUM GRATIĀ
------------------------------------------------------------------------------
""".format(
    PROGRAM
)

# Record separator
DICTIONARY_SEPARATOR = "␞"

STDIN = sys.stdin.buffer


class Cli:
    """Main CLI parser"""

    def __init__(self):
        """
        Constructs all the necessary attributes for the Cli object.
        """
        self.pyargs = None
        self.EXIT_OK = 0
        self.EXIT_ERROR = 1
        self.EXIT_SYNTAX = 2

    def make_args(self):
        """make_args"""
        parser = argparse.ArgumentParser(
            prog=PROGRAM,
            description=DESCRIPTION,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            epilog=__EPILOGUM__,
        )

        parser.add_argument("input", help="path to CSV file on disk. Use - for stdin")

        parser.add_argument(
            "--input-delimiter",
            help="The input delimiter",
            dest="in_delimiter",
            default=",",
            required=False,
            nargs="?",
        )

        parser.add_argument(
            "--input-encoding",
            help="The input encoding",
            dest="in_encoding",
            default="utf-8",
            required=False,
            nargs="?",
        )

        parser.add_argument(
            "--input-fieldnames",
            help="If the input CSV does not have a header, specify here. "
            "Use | as separator (if a field de de facto have |, then use \|). "
            "Example: --input-fieldnames='field with \| on it|another field'",
            dest="in_fieldnames",
            # default="utf-8",
            required=False,
            nargs="?",
        )

        parser.add_argument(
            "--dict-target-key",
            help="Field name to represent the primary key to convert data "
            "Defaults to first column. Example: "
            "--dict-target-key='id'",
            dest="dict_target",
            required=False,
            nargs="?",
        )

        parser.add_argument(
            "--dict-source-key",
            help="Field name to be used as source to convert data to target key. "
            "If undefined, defaults to all fields which are not the "
            "--dict-target-key. Example: "
            "--dict-source-key='name' --dict-source-key='title' ",
            dest="dict_sources",
            action="append",
            required=False,
            nargs="?",
        )

        parser.add_argument(
            "--transform-uppercase",
            help="Force all source values to UPPERCASE",
            dest="t_uppercase",
            action="store_true",
        )

        parser.add_argument(
            "--transform-lowercase",
            help="Force all source values to lowercase",
            dest="t_lowercase",
            action="store_true",
        )

        parser.add_argument(
            "--transform-no-latin-accents",
            help="Remove some diacrilics of latin script",
            dest="t_nolatinaccents",
            action="store_true",
        )

        parser.add_argument(
            "--logfile",
            help="Path to a file to log warnings and other information",
            dest="logfile",
            nargs="?",
        )

        parser.add_argument(
            "--ignore-warnings",
            help="Ignore some errors (duplicated key / ambiguous results)",
            dest="ignore_warnings",
            action="store_true",
        )

        return parser.parse_args()

    def execute_cli(self, pyargs, stdin=STDIN, stdout=sys.stdout, stderr=sys.stderr):
        # @TODO implement strict run (fail if repeated source falues)
        # @TODO implement UPPER, lower and remove-accents options

        f_warnings = sys.stderr
        if pyargs.logfile:
            f_logfile = open(pyargs.logfile, "w", encoding="utf-8")
            f_warnings = f_logfile

        # @TODO stdin does not yet allow non UTF8 customization (will pass as it is)
        # @see https://stackoverflow.com/questions/5004687
        with open(pyargs.input, "r", encoding=pyargs.in_encoding) if len(
            pyargs.input
        ) > 1 else sys.stdin as csvfile:
            if pyargs.in_fieldnames:
                _in_fieldnames = re.split(r"(?<!\\)\|", pyargs.in_fieldnames)
                # _in_fieldnames = re.split(r"", pyargs.in_fieldnames)

                reader = csv.DictReader(
                    csvfile, fieldnames=_in_fieldnames, delimiter=pyargs.in_delimiter
                )
            else:
                reader = csv.DictReader(csvfile, delimiter=pyargs.in_delimiter)

            # reader = csv.DictReader(csvfile, delimiter=pyargs.in_delimiter)

            firstline = next(reader)
            # print(firstline)

            _fieldnames = firstline.keys()
            # print(_fieldnames)

            if pyargs.dict_target:
                dict_target = pyargs.dict_target
            else:
                # dict_target = _fieldnames[0]
                dict_target = list(_fieldnames).pop(0)

            if pyargs.dict_sources:
                dict_sources = pyargs.dict_sources
            else:
                _temp = list(_fieldnames)
                _temp.remove(dict_target)
                dict_sources = _temp

            writer = csv.writer(
                sys.stdout, delimiter=DICTIONARY_SEPARATOR, quoting=csv.QUOTE_MINIMAL
            )

            outdict = {}
            # @TODO This part is a bit duplicated. Could be better.
            # #     Remove duplicated code
            for item in dict_sources:
                if not firstline[item] or len(firstline[item]) == 0:
                    continue

                _v = firstline[item].strip()
                _k = firstline[dict_target].strip()

                if pyargs.t_nolatinaccents:
                    _v = _v.lower()
                    # Obviously incomplete
                    _v = re.sub(r"[àáâãäå]", "a", _v)
                    _v = re.sub(r"[èéêë]", "e", _v)
                    _v = re.sub(r"[ìíîï]", "i", _v)
                    _v = re.sub(r"[òóôõö]", "o", _v)
                    _v = re.sub(r"[ñ]", "n", _v)
                    _v = re.sub(r"[ç]", "c", _v)

                if pyargs.t_lowercase:
                    _v = _v.lower()
                elif pyargs.t_uppercase:
                    _v = _v.upper()

                if _k in outdict and outdict[_k] != _v and not pyargs.ignore_warnings:
                    print(f"{_k} repeated", file=f_warnings)
                    continue

                outdict[_k] = _v

            for row in reader:
                for item in dict_sources:
                    if not row[item] or len(row[item]) == 0:
                        continue

                    _v = row[item].strip()
                    _k = row[dict_target].strip()

                    if pyargs.t_nolatinaccents:
                        _v = _v.lower()
                        # Obviously incomplete
                        _v = re.sub(r"[àáâãäå]", "a", _v)
                        _v = re.sub(r"[èéêë]", "e", _v)
                        _v = re.sub(r"[ìíîï]", "i", _v)
                        _v = re.sub(r"[òóôõö]", "o", _v)
                        _v = re.sub(r"[ñ]", "n", _v)
                        _v = re.sub(r"[ç]", "c", _v)

                    if pyargs.t_lowercase:
                        _v = _v.lower()
                    elif pyargs.t_uppercase:
                        _v = _v.upper()

                    if (
                        _k in outdict
                        and outdict[_k] != _v
                        and not pyargs.ignore_warnings
                    ):
                        print(f"{_k} repeated", file=f_warnings)
                        continue

                    outdict[_k] = _v

                    # outdict[firstline[item].strip()] = firstline[dict_target].strip()
                # writer.writerow(row)

            outdict_sorted = dict(sorted(outdict.items()))

            # print(outdict)
            # outdict

            for key, value in outdict_sorted.items():
                writer.writerow([key, value])

            # # writer.writeheader()
            # writer.writerow(firstline)

            # for row in reader:
            #     writer.writerow(row)

        if pyargs.logfile:
            f_logfile.close()

        return self.EXIT_OK


def exec_from_console_scripts():
    main = Cli()
    args = main.make_args()
    main.execute_cli(args)


if __name__ == "__main__":
    cli_2600 = Cli()
    args = cli_2600.make_args()
    # pyargs.print_help()

    # args.execute_cli(args)
    cli_2600.execute_cli(args)
